/*

 ideserv.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>
 Win32 port by
  Josef Soucek <josef.soucek@ide64.org>
 Serial support by
  Borsanyi Attila <bigfoot@axelero.hu>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"
#include <stdio.h>
#ifndef WIN32		//*NIX
#include <errno.h>
#include <sys/time.h>
#include <pwd.h>
#include <grp.h>
#include <sys/resource.h>
#include <sys/wait.h>
#else			//WIN32
#include <windows.h>
#include <errno.h>
#include "resource.h"
#include "timeval.h"
#include "sleep.h"
#define lstat(x,y) stat(x,y)
#define S_ISLNK(x) 0
#endif 			//ALL
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>
#include <sys/stat.h>
#include <stdlib.h>
#include "my_getopt-1.4/getopt.h"
#include <sys/types.h>
#include "path.h"
#include "partition.h"
#include "driver.h"
#include "usb.h"
#include "eth.h"
#ifndef WIN32
#include "fifo.h"
#endif
#include "rs232.h"
#ifndef OSX
#include "x1541.h"
#include "pc64.h"
#endif

static const struct io_ops *io_ops;
static FILE *lf;
static struct buffer_s {
    unsigned char *data;
    unsigned short pointer;
    unsigned short length;
    FILE *file;
    int fd;
    int mode;
    unsigned int filesize;
} buff[16];
static char errorbuff[256];

#include "crc8.h"
#include "ideserv.h"
static struct arguments_s arguments;

#define CM_CLOSED 0
#define CM_DIR 1
#define CM_FILE 2
#define CM_COMPAT 3

#ifdef WIN32
/* Prototypes */
static INT_PTR CALLBACK MyDlgProc(HWND,UINT,WPARAM,LPARAM);
static long WINAPI PCLinkLoop(long);
/* Global variable */
static NOTIFYICONDATA g_notifyIconData;
#endif

static const char *argp_program_version="ideservd 0.28b 20140929";

static void terminate() {
    if (arguments.mode != M_ETHERNET) io_ops->shutdown();
    if (lf) {
	fprintf(lf,"Terminated.\n");
	fclose(lf);
    }
    exit(0);
}

static void seterror(int c,int i1) {
    char *msg="UNKNOWN ERROR";

    switch (c) {
    case 00:msg="OK";break;
    case 01:msg="FILES SCRATCHED";break;
    case 02:msg="PARTITION SELECTED";break;
    case 23:if (i1==0) msg="READ ERROR"; else msg="CRC ERROR";i1=0;break;
    case 26:msg="WRITE PROTECT ON";break;
    case 27:msg="ACCESS DENIED";break;
    case 31:msg="UNKNOWN COMMAND";break;
    case 30:
    case 33:
    case 34:msg="SYNTAX ERROR";break;
    case 39:msg="PATH NOT FOUND";break;
    case 62:msg="FILE NOT FOUND";break;
    case 63:msg="FILE EXISTS";break;
    case 72:msg="PARTITION FULL";break;
    case 73:msg="IDESERV 0.27B PCLINK";break;
    case 77:msg="SELECTED PARTITION ILLEGAL";break;
    }
    sprintf(errorbuff,"%02d, %s,%03d,000,000,000",c,msg,i1);
}

static int send_trailer(int usecrc) {
    if (arguments.mode != M_ETHERNET) {
        if (usecrc) {
            if (io_ops->sendb(crc_get())) return 1;
        }
        if (io_ops->sendb(0x5a)) return 1;
    }
    return io_ops->flush();
}

static int check_trailer(char *mode, int usecrc) {
    int fr;
    if (arguments.mode == M_ETHERNET) return 0;
    if (usecrc) {
        if (io_ops->getb() < 0) return 1;
        if (crc_get()) {
	    seterror(23, 1);
            fprintf(lf, "CRC error while %s\n", mode);
            io_ops->getb();
            return 1;
        }
    }
    fr = io_ops->getb();if (fr < 0) return 1;
    if (fr != 0x5a) {
	fprintf(lf, "Frame error %02x while %s\n", fr, mode);
	return 1;
    }
    return 0;
}

volatile int tout;
static void timeout() {
    fprintf(lf,"Timeout.\n");
    io_ops->clean();
    tout=1;
}

void errtochannel15(int i) {
    if (i)
        switch (errno) {
        case EISDIR:
        case ENOTDIR:
        case ENOENT:seterror(62,0);break;
        case EEXIST:seterror(63,0);break;
        case EPERM:
        case EACCES:seterror(27,0);break;
        case ENOTEMPTY:seterror(1,0);break;
        case ENOSPC:seterror(72,0);break;
        case EROFS:seterror(26,0);break;
        case EIO:seterror(23,0);break;
        default:seterror(80,errno);
        }
    else seterror(0,0);
}

static void do_chdir(const c64c_t *cmd) {
    struct dirent *ep;
    c64c_t name[17];
    const c64c_t *outname;
    DIR *dp;
    char outpath[1020];
    int found=0;
    partition_t outpart = 0;

    if (!strchr((char *)cmd, ':')) {
        if (!strcmp((char *)cmd, "_")) cmd=(c64c_t *)":_";
        else if (cmd[0] != '/') { seterror(30,0); goto vege2; }
        else cmd++;
    }

    if (!(found=resolv_path(cmd, outpath, &outname, &outpart, lf, arguments.relaxfiletypes))) {seterror(39,0);goto vege2;}

    if (!outname[0]) goto vege2;
    if (!strcmp((const char *)outname,"_") || !strcmp((const char *)outname, "..")) {
        char *o = strrchr(outpath, '/');
        if (o) *o=0; else outpath[0]=0;
        goto vege2;
    }
    if (!strcmp((const char *)outname, ".")) goto vege2;
    found = 0;
    convertfilename(outname, ',', name, NULL, NULL);

    dp=opendir(outpath[0]?outpath:".");
    if (!dp) {errtochannel15(1);fprintf(lf,"Couldn't open the directory \"%s\".\n",outpath);return;}

    initdirentry(0);
    while ((ep=readdir(dp)))
    {
        DIRECTORY_ENTRY_T dirent;
        char lname2[2020];

        strcpy(lname2,outpath);
        if (outpath[0]) strncat(lname2,"/",999);
        strncat(lname2,ep->d_name,999);

        if (createdirentry(ep->d_name, lname2, &dirent, arguments.relaxfiletypes)) continue;

        if ((dirent.attrib & A_ANY)!=A_DIR) continue;

        dirent.filesize[0]=0;
        if (!matchname(dirent.name, name)) continue;

        strcpy(outpath, lname2);
        found=1;
        break;
    }
    closedir(dp);
    initdirentry(1);
vege2:
    if (found) {
        partition_set_path(outpart, outpath);
        seterror(0,0);
    } else seterror(39,0);
    fprintf(lf,"Chdir to \"%s\".\n", outpath[0]?outpath:"/");
}

static void do_mkdir(const c64c_t *cmd) {
    struct dirent *ep;
    DIR *dp;
    char outpath[1020],lname[1000];
    int found=0;
    c64c_t name[17];
    const c64c_t *outname;

    if (!strchr((char *)cmd, ':')) {seterror(30,0);found=0;outpath[0]=0;goto vege2;}

    found=0;
    if (!resolv_path(cmd, outpath, &outname, NULL, lf, arguments.relaxfiletypes)) {seterror(39,0);goto vege2;}

    convertfilename(outname, ',', name, NULL, NULL);

    dp=opendir(outpath[0]?outpath:".");
    if (!dp) {errtochannel15(1);fprintf(lf,"Couldn't open the directory \"%s\".\n",outpath);return;}

    initdirentry(0);
    while ((ep=readdir(dp)))
    {
        DIRECTORY_ENTRY_T dirent;
        char lname2[2020];

        strncpy(lname,ep->d_name,999);
        strcpy(lname2,outpath);
        if (outpath[0]) strncat(lname2,"/",999);
        strncat(lname2,ep->d_name,999);

        if (createdirentry(ep->d_name, lname2, &dirent, arguments.relaxfiletypes)) continue;

        if ((dirent.attrib & A_ANY)!=A_DIR) continue;

        dirent.filesize[0]=0;
        if (!matchname(dirent.name, name)) continue;

        found=1;
        break;
    }
    closedir(dp);
    initdirentry(1);

    if (!found) {
        int f;
        for (f = 0; name[f] && f < 16; f++) {
            lname[f]=c64toascii(name[f]);
        }
        lname[f]=0;
    }

    if (outpath[0]) strcat(outpath,"/");
    strcat(outpath,lname);

    if (found)
        seterror(63,0);
    else {
        int f=0;

        while (lname[f]) {
            if (lname[f]=='*' || lname[f]=='?' || lname[f]==':' || lname[f]=='=') {
                seterror(33,0);goto vege2;
            }
            f++;
        }
        errtochannel15(mkdir(outpath,0777));
    }
vege2:
    if (!name[0]) {
        seterror(34,0);
    }
    fprintf(lf,"Mkdir \"%s\".\n",outpath[0]?outpath:"/");
}

static void do_scratch(const c64c_t *cmd, int smode) {
    char outpath[1000];
    struct dirent *ep;
    DIR *dp;
    int found=0;
    c64c_t name[17], type[4]={'*',0};
    const c64c_t *outname;

    if (!strchr((char *)cmd, ':')) {seterror(30,0);found=0;outpath[0]=0;goto vege;}

    found=0;
    if (!resolv_path(cmd, outpath, &outname, NULL, lf, arguments.relaxfiletypes)) {seterror(39,0);goto vege;}

    convertfilename(outname, '=', name, type, NULL);

    dp=opendir(outpath[0]?outpath:".");
    if (!dp) {errtochannel15(1);fprintf(lf,"Couldn't open the directory \"%s\".\n",outpath);return;}

    initdirentry(0);
    while ((ep=readdir(dp)))
    {
        DIRECTORY_ENTRY_T dirent;
        char lname2[2020];

        strcpy(lname2,outpath);
        if (outpath[0]) strncat(lname2,"/",999);
        strncat(lname2,ep->d_name,999);

        if (createdirentry(ep->d_name, lname2, &dirent, arguments.relaxfiletypes)) continue;

        if (smode) {
            if ((dirent.attrib & (A_ANY | A_CLOSED))!=(A_DIR | A_CLOSED)) continue;
        }
        else
            if ((dirent.attrib & A_ANY)==A_DIR) continue;

        if (!(dirent.attrib & A_DELETEABLE)) continue;

        dirent.filesize[0]=0;
        if (!matchname(dirent.name, name)) continue;
        if (!smode) {
            dirent.time[0]=0;
            if (!matchname(dirent.filetype,type)) continue;
        }

        found=1;
        if (smode) {
            fprintf(lf,"Rmdir \"%s\"\n",lname2);
            errtochannel15(rmdir(lname2));
        } else {
            errtochannel15(unlink(lname2));
            fprintf(lf,"Rm \"%s\"\n",lname2);
        }
    }
    closedir(dp);
    initdirentry(1);
    if (!found) seterror(1,0);
vege:
    if (!found) {
        if (smode) {
            fprintf(lf,"Rmdir \"%s\"\n",outpath);
        } else {
            fprintf(lf,"Rm \"%s\"\n",outpath);
        }
    }

    if (!name[0]) {
        seterror(34,0);
    }
}

static void commandchannel(const c64c_t *s) {
    if (s[0]=='C' && s[1]=='D') {
        do_chdir(s+2);
    } else if (s[0]=='M' && s[1]=='D') {
        do_mkdir(s);
    } else if (s[0]=='R' && s[1]=='D') {
        do_scratch(s,1);
    } else if (s[0]=='S' && s[1]!='-') {
        do_scratch(s,0);
    } else if (s[0]=='C' && s[1]==0xd0) {
        fprintf(lf,"Change to partition %d\n",s[2]);
        seterror(partition_select(s[2]),s[2]);
    } else if (s[0]=='C' && s[1]=='P') {
        partition_t part = 0;
        int i = 2;
        if (!s[2]) part = partition_get_current(); else {
            while ((s[i] ^ 0x30) < 10) part = part * 10 + (s[i++] ^ 0x30);
        }
        fprintf(lf,"Change to partition %d\n",part);
        seterror(partition_select(part), part);
    } else if (s[0]=='U' && ((s[1] & 0x0f)==9 || (s[1] & 0x0f)==10)) {
        seterror(73,0);
        fprintf(lf,"Identify device\n");
    } else if (s[0]=='I') {
        seterror(0,0);
        fprintf(lf,"Initialize\n");
    } else if (s[0]=='V') {
        seterror(0,0);
        fprintf(lf,"Validate\n");
    } else if (s[0]=='M' && s[1]=='-' && s[2]=='R') { 
        int i,j,k;
        j=s[5];if (j>50) j=50;
        k=s[3];
	for (i=0;i<j;i++) errorbuff[i]="IDE64 CARTRIDGE "[k++ & 15];
	errorbuff[i]=0;
        fprintf(lf,"Memory read $%04x %02x\n", s[3] | (s[4] << 8), s[5]);
    } else if (s[0]) {
        seterror(31,0);
    } else fprintf(lf,"\n");
}

static int statuserror(int usecrc) {
    c64c_t cmd[256];
    unsigned char fp;
    int r;

    if (io_ops->getb() < 0) return 1;
    fp = 0;
    while (fp<255) {
        r = io_ops->getb();
        if (r < 0) return 1;
        cmd[fp] = r;
        if (!cmd[fp]) break;
        fp++;
    }
    cmd[fp] = 0;

    if (check_trailer("status", usecrc)) return 1;

    alarm(0);

    if (arguments.debug) {
        int i=0;
        fprintf(lf, "Command:\n");
        while (cmd[i]) {
            fprintf(lf, " %02x", cmd[i++]);
            if (!(i & 15)) fputc('\n', lf);
        }
        if (i & 15) fputc('\n', lf);
    }
    commandchannel(cmd);

    crc_clear();
    fp = strlen(errorbuff);
    if (io_ops->sendb(fp)) return 1;
    if (arguments.mode == M_ETHERNET && io_ops->sendb(fp)) return 1;
    if (io_ops->sendbytes(fp, (unsigned char *)errorbuff)) return 1;
    if (send_trailer(usecrc)) return 1;
    return 0;
}

#include "compat.h"

static int openfile(int usecrc) {
    c64c_t cmd[256];
    unsigned char status=0x80;
    unsigned char mode='R';
    unsigned char channel,fp;
    unsigned long long l=0;
    unsigned int length=0;
    struct buffer_s *buffer;
    int r;

    r = io_ops->getb();
    if (r < 0) return 1;
    channel=r & 0x0f;
    buffer=&buff[channel];

    fp=0;
    while (fp < 255) {
        r = io_ops->getb();
        if (r < 0) return 1;
        cmd[fp] = r;
        if (!r) break;
        fp++;
    }
    cmd[fp]=0;

    if (check_trailer("open",usecrc)) return 1;

    alarm(0);

    if (arguments.debug) {
        int i=0;
        fprintf(lf, "Command:\n");
        while (cmd[i]) {
            fprintf(lf, " %02x", cmd[i++]);
            if (!(i & 15)) fputc('\n', lf);
        }
        if (i & 15) fputc('\n', lf);
    }
    {
        if (buffer->file) {
            fclose(buffer->file);
            buffer->file=NULL;
        }
        buffer->mode=CM_CLOSED;

        if (cmd[0] == '$') {
            unsigned char dirline[32];
            struct dirent *ep;
            DIR *dp;
            int sum=0,partition;
            char outpath[1000];
            buffer->mode=CM_DIR;
            buffer->length=buffer->pointer=0;
            c64c_t name[17], type[4]={'*',0};
            const c64c_t *outname;
            partition_t outpart = 0;

	    partition=!strncmp((char *)&cmd[1],"=P",2);

            resolv_path(cmd+1, outpath, &outname, &outpart, lf, arguments.relaxfiletypes);
            fprintf(lf,"Open [%d]: \"$%s\"\n", channel, outpath);
            convertfilename(outname, '=', name, type, NULL);

            if (!name[0]) strcpy((char *)name, "*");

            if (channel==1) {
                status = 0x80 | 70;goto vege; //no channel
            } else if (channel>1) {
                dp=opendir(outpath[0]?outpath:".");
                if (!dp) {
                    fprintf(lf,"Couldn't open the directory \"%s\".\n",outpath);
                    status = 0x80 | 62;goto vege;
                }
                if (!buffer->data) {
                    buffer->data=malloc(65536);
                    if (!buffer->data) exit(1);//out of memory?!
                }
                buffer->data[buffer->length++]='I';

                memset(&dirline[0],32,16);
                memset(&dirline[16],0,8);
                dirline[24]=A_DIR;
                memcpy(&dirline[25],"DIR",3);
                memset(&dirline[28],0,4);
                for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];

                initdirentry(0);
                while ((ep=readdir(dp)))
                {
                    char lname2[2020];

                    strcpy(lname2,outpath);
                    if (outpath[0]) strncat(lname2,"/",999);
                    strncat(lname2,ep->d_name,999);

                    if (createdirentry(ep->d_name, lname2, (DIRECTORY_ENTRY_T*)dirline, arguments.relaxfiletypes)) continue;

                    for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];
                }
                initdirentry(1);
                memset(buffer->data + buffer->length, 0, (-buffer->length) & 511);

                errtochannel15(closedir(dp));
                status=0x80;//ok
            } else if (partition) {
                int i, n=0;
                memcpy(&dirline[0x00],"\1\4\1\1\377\0\22\42",8);
                memcpy(&dirline[0x08],"IDESERVD 0.27B  ",16);
#ifdef WIN32
                memcpy(&dirline[0x18],"\42\40WIN32\0",8);
#else
                memcpy(&dirline[0x18],"\42\40LINUX\0",8);
#endif
                if (!buffer->data) {
                    buffer->data=malloc(65536);
                    if (!buffer->data) exit(1);//out of memory?!
                }
                for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];
                for (i=1;i<255;i++) {
                    c64c_t *name = partition_get_name(i);
                    if (!name) continue;
                    dirline[0x00]=dirline[0x01]=1;
                    dirline[0x02]=i;
                    dirline[0x03]=0;
                    memcpy(&dirline[0x04],"   \42\42                 NET  \0",28);
                    memcpy(&dirline[0x08], name, strlen((char *)name));
                    dirline[0x08 + strlen((char *)name)] = '"';
                    for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];
                    n++;
                }
                dirline[0x00]=dirline[0x01]=1;
                dirline[0x02]=n;
                dirline[0x03]=0;
                memcpy(&dirline[0x04],"PARTITIONS.              \0\0\0",28);
                for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];
                memset(buffer->data + buffer->length, 0, (-buffer->length) & 511);
                errtochannel15(0);
                status=0x80;//ok
            } else {

                dp=opendir(outpath[0]?outpath:".");

                if (!dp) {
                    fprintf(lf,"Couldn't open the directory \"%s\".\n",outpath);
                    status = 0x80 | 62;goto vege;
                }

                memcpy(&dirline[0x00],"\1\4\1\1\1\0\22\42",8);
                dirline[0x04] = outpart ? outpart : partition_get_current();
                memset(&dirline[0x08],32,16);
#ifdef WIN32
                memcpy(&dirline[0x18],"\42\40WIN32\0",8);
#else
                memcpy(&dirline[0x18],"\42\40LINUX\0",8);
#endif
                if (!buffer->data) {
                    buffer->data=malloc(65536);
                    if (!buffer->data) exit(1);//out of memory?!
                }
                for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];

                initdirentry(0);
                while ((ep=readdir(dp)))
                {
                    DIRECTORY_ENTRY_T dirent;
                    int offs;
                    unsigned int n;
                    char lname2[2020];

                    strcpy(lname2,outpath);
                    if (outpath[0]) strncat(lname2,"/",999);
                    strncat(lname2,ep->d_name,999);

                    if (createdirentry(ep->d_name, lname2, &dirent, arguments.relaxfiletypes)) continue;

                    if (dirent.pointer[0] & 0x80) continue;
                    n=(dirent.filesize[0]+(dirent.filesize[1] << 8)+(dirent.filesize[2] << 16)+(dirent.filesize[3] << 24)+255) >> 8;
                    if (n>65535) n=65535;

                    dirent.filesize[0]=0;
                    if (!matchname(dirent.name, name)) continue;
                    dirent.time[0]=0;
                    if (!matchname(dirent.filetype,type)) continue;

                    dirline[0x00]=dirline[0x01]=1;
                    dirline[0x02]=n;
                    dirline[0x03]=n >> 8;
                    sum+=n;
                    memset(&dirline[0x04],32,27);
                    offs=7;
                    if (n>9) offs--;
                    if (n>99) offs--;
                    if (n>999) offs--;
                    dirline[offs++]=0x22;
                    n=0;while (dirent.name[n] && n<16) dirline[offs++]=dirent.name[n++];
                    dirline[offs++]=0x22;
                    offs+=(16-n);

                    dirline[offs++]=(dirent.attrib & A_CLOSED)?' ':'*';

                    n=0;while (n<3 && dirent.filetype[n]) {dirline[offs+n]=dirent.filetype[n];n++;}
                    if ((dirent.attrib & (A_DELETEABLE | A_WRITEABLE))!=(A_DELETEABLE | A_WRITEABLE)) dirline[offs+3]='<';

                    for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];
                    if (sum>65535) sum=65535;
                }
                initdirentry(1);

                dirline[0x00]=dirline[0x01]=1;
                dirline[0x02]=sum;
                dirline[0x03]=sum >> 8;
                memcpy(&dirline[0x04],"BLOCKS USED.             \0\0\0",28);

                for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];
                memset(buffer->data + buffer->length, 0, (-buffer->length) & 511);
                errtochannel15(closedir(dp));
                status=0x80;//ok
            }
            length=buffer->length;
        } else {
            c64c_t name[17], type[4]={'*',0};
            const c64c_t *outname;
            unsigned char overwrite;
            char lname[1000],outpath[1020];
            struct dirent *ep;
            DIR *dp;
            int found=0;
            int b;

            overwrite=0;
            for (b=0;cmd[b];b++) if (cmd[b]==':') {
                if (cmd[0]=='@') overwrite=1;
                break;
            }

            resolv_path(cmd, outpath, &outname, NULL, lf, arguments.relaxfiletypes);
            if (channel==1) strcpy((char *)type,"PRG");
            convertfilename(outname, ',', name, type, &mode);

            dp=opendir(outpath[0]?outpath:".");
            if (!dp) {
                fprintf(lf,"Couldn't open the directory \"%s\".\n",outpath);
                status = 0x80 | 62;goto vege;
            }

            initdirentry(0);
            while ((ep=readdir(dp)))
            {
                DIRECTORY_ENTRY_T dirent;
                char lname2[2020];

                strncpy(lname,ep->d_name,999);
                strcpy(lname2,outpath);
                if (outpath[0]) strncat(lname2,"/",999);
                strncat(lname2,ep->d_name,999);

                if (createdirentry(ep->d_name, lname2, &dirent, arguments.relaxfiletypes)) continue;

                if ((dirent.attrib & (A_ANY | A_CLOSED))!=(A_CLOSED | A_NORMAL)) continue;

                length = dirent.filesize[0] + dirent.filesize[1] * 256 + dirent.filesize[2] * 65536 + dirent.filesize[3] * 65536 * 256;

                dirent.filesize[0]=0;
                if (!matchname(dirent.name, name)) continue;
                dirent.time[0]=0;
                if (!matchname(dirent.filetype,type)) continue;

                found=1;
                break;
            }
            closedir(dp);
            initdirentry(1);

            if (!found) {
                int f=0,g=0;
                while (name[f]) {
                    if (f<16) {
                        lname[f]=c64toascii(name[f]);
                        f++;
                    }
                }
		if (!arguments.relaxfiletypes || memcmp(type,"PRG",3)) {
            	    lname[f++]=',';
            	    while (type[g]) {
                	if (g<3) {
                    	    lname[f]=c64toascii(type[g]);
                    	    f++;g++;
                	}
            	    }
		}
                lname[f]=0;
            }

            if (channel==0) mode='R';
            if (channel==1) mode='W';

            if (outpath[0]) strcat(outpath,"/");
            strcat(outpath,lname);

            fprintf(lf,"Open [%d]: \"%s\" %c\n",channel,outpath,mode);

            if (!name[0]) {
                status=0x80 | 34;goto vege;//no filename
            }

            if (mode=='W') {
                int f=0;
                length=0;

                while (lname[f]) {
                    if (lname[f]=='*' || lname[f]=='?' || lname[f]==':' || lname[f]=='=') {
                        status=0x80 | 33;goto vege;//illegal characters
                    }
                    f++;
                }

                if (found && !overwrite) {
                    status=0x80 | 63;//exists
                } else {
#ifndef WIN32
                    if (overwrite) buffer->fd=open(outpath, O_CREAT|O_RDWR|O_TRUNC, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
                    else buffer->fd=open(outpath, O_CREAT|O_EXCL|O_RDWR, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
#else
                    if (overwrite) buffer->fd=open(outpath, O_CREAT|O_RDWR|O_TRUNC|O_BINARY, S_IRUSR|S_IWUSR);
                    else buffer->fd=open(outpath, O_CREAT|O_EXCL|O_RDWR|O_BINARY, S_IRUSR|S_IWUSR);
#endif
                    if (buffer->fd<0) errtochannel15(1); else {
                        status=0x80;
                        buffer->file=fdopen(buffer->fd,"wb+");
                        buffer->mode=CM_FILE;
                        buffer->filesize=0;
                    }
                }
            } else {
                if (!found) {
                    status=0x80 | 62;//not found
                } else {
        	    if (mode=='R') {

#ifndef WIN32
                	buffer->fd=open(outpath, O_RDONLY, 0);
#else
                	buffer->fd=open(outpath, O_RDONLY|O_BINARY, 0);
#endif
                	if (buffer->fd<0) errtochannel15(1); else {
                	    status=0x80;//ok
                    	    buffer->file=fdopen(buffer->fd, "rb");
                    	    buffer->mode=CM_FILE;
                            buffer->filesize=length;
                	}
            	    }
        	    if (mode=='A' || mode=='M') {
#ifndef WIN32
                	buffer->fd=open(outpath, O_RDWR, 0);
#else
                	buffer->fd=open(outpath, O_RDWR|O_BINARY, 0);
#endif
                	if (buffer->fd<0) errtochannel15(1); else {
                	    status=0x80;//ok
                    	    buffer->file=fdopen(buffer->fd, "rb+");
                            buffer->mode=CM_FILE;
                            buffer->filesize=length;
                	}
            	    }
                }
            }
        }
    }
vege:
    {
        unsigned char buf[6];
        crc_clear();
        if (arguments.mode == M_ETHERNET) {
            buf[0] = length;
            buf[1] = length >> 8;
            buf[2] = length >> 16;
            buf[3] = length >> 24;
            buf[4] = mode;
            buf[5] = status;
        } else {
            buf[0] = status;
            buf[1] = mode;
            buf[2] = length;
            buf[3] = length >> 8;
            buf[4] = length >> 16;
            buf[5] = length >> 24;
        }
        if (io_ops->sendbytes(sizeof(buf), buf)) return 1;
        if (send_trailer(usecrc)) return 1;
    }
    return 0;
}

static int closefile(int usecrc) {
    unsigned char channel;
    unsigned int length;
    struct buffer_s *buffer;
    unsigned char buf[5];

    if (io_ops->getbytes(sizeof(buf), buf)) return 1;
    channel=buf[0] & 0x0f;
    length=(((((buf[4] << 8) | buf[3]) << 8) | buf[2]) << 8) | buf[1];
    if (check_trailer("close",usecrc)) return 1;
    buffer=&buff[channel];

    alarm(0);

    fprintf(lf,"Close [%d] %d byte(s)\n",channel, length);
    if (buffer->file) {
        if (buffer->mode==CM_FILE && length >= buffer->filesize) {
            fflush(buffer->file);
            ftruncate(buffer->fd,length);
        }
	fclose(buffer->file);
    }
    buffer->mode=CM_CLOSED;buffer->file=NULL;
    if (buffer->data) {
        free(buffer->data);
        buffer->data=NULL;
    }
    crc_clear();
    if (io_ops->sendb(0x80)) return 1;
    if (arguments.mode == M_ETHERNET && io_ops->sendb(0x80)) return 1;
    if (send_trailer(usecrc)) return 1;
    return 0;
}

static int readfile(int usecrc) {
    unsigned char channel, sectors;
    unsigned short int j;
    struct timeval start,end;
    unsigned long long l=0,l2;
    unsigned int address;
    struct buffer_s *buffer;
    int fr;
    unsigned char buf[5];

    if (io_ops->getbytes(sizeof(buf), buf)) return 1;
    channel=buf[0] & 0x0f;
    address=(((buf[3] << 8) | buf[2]) << 8) | buf[1];
    sectors=buf[4];
    if (check_trailer("read",usecrc)) return 1;
    if (arguments.mode == M_ETHERNET) {
        if (sectors>2) sectors=2;
    } else {
        if (sectors>128) sectors=128;
    }
    buffer=&buff[channel];

    alarm(0);

    fprintf(lf,"Read [%d] %06x %d sector(s)",channel,address, sectors);
    if (buffer->mode==CM_CLOSED) {
        fprintf(lf," channel not open\n");
        crc_clear();
        if (io_ops->sendb(0x80 | 70)) return 1;
        if (arguments.mode == M_ETHERNET && io_ops->sendb(0x80 | 70)) return 1;
        if (send_trailer(usecrc)) return 1;
        return 0;
    }
    if (buffer->mode!=CM_DIR && (buffer->mode!=CM_FILE || !buffer->file)) {
        fprintf(lf," not readable.\n");
        crc_clear();
        if (io_ops->sendb(0x80 | 70)) return 1;
        if (arguments.mode == M_ETHERNET && io_ops->sendb(0x80 | 70)) return 1;
        if (send_trailer(usecrc)) return 1;
        return 0;
    }

    if (buffer->mode==CM_FILE) {
	unsigned long itt;
	fseek(buffer->file, address << 8, SEEK_SET);
        if (!buffer->data) {
            buffer->data=malloc(65536);
            if (!buffer->data) exit(1);//out of memory?!
        }
        itt=fread(buffer->data,1,sectors*512,buffer->file);

	if (itt<sectors*512) memset(buffer->data+itt,0,sectors*512-itt);
	j=0;
    } else {
        j=buffer->pointer;
    }

    if (arguments.mode == M_ETHERNET) {
        crc_clear();
        if (io_ops->sendbytes(sectors * 512, buffer->data + j)) return 1;
        if (io_ops->sendb(0x80)) return 1;
        if (io_ops->sendb(0x80)) return 1;
        if (send_trailer(usecrc)) return 1;
        buffer->pointer+=sectors*512;
    } else {
        crc_clear();
        if (io_ops->sendb(0x80)) return 1;
        if (send_trailer(usecrc)) return 1;
        gettimeofday(&start,NULL);
        while (sectors) {
            if (arguments.mode == M_RS232 || arguments.mode == M_RS232S) {
                fr = io_ops->getb();if (fr < 0) return 1;
                if (fr != 0x5a) {
                    fprintf(lf, " frame error %02x\n", fr);
                    return 1;
                }
            }
            crc_clear();
            if (io_ops->sendbytes(512,buffer->data+j)) return 1;
            j+=512;
            if (io_ops->sendb(0x80)) return 1;
            if (send_trailer(usecrc)) return 1;
            sectors--;buffer->pointer=j;l+=512;
        }
        gettimeofday(&end,NULL);
        l2=((end.tv_sec-start.tv_sec)*1000000+end.tv_usec-start.tv_usec);
        if (l2) {
            l=l*1000000/l2;
            fprintf(lf," %lld bytes/seconds\n",l);
        }
    }
    return 0;
}


static int writefile(int usecrc) {
    unsigned char channel, sectors;
    struct timeval start,end;
    unsigned long long l=0,l2;
    unsigned int address;
    struct buffer_s *buffer;
    int fr;
    unsigned char buf[5];

    if (io_ops->getbytes(sizeof(buf), buf)) return 1;
    channel=buf[0] & 0x0f;
    address=(((buf[3] << 8) | buf[2]) << 8) | buf[1];
    sectors=buf[4];
    if (check_trailer("write",usecrc)) return 1;
    if (sectors>128) sectors=128;
    buffer=&buff[channel];

    alarm(0);

    fprintf(lf,"Write [%d] %06x %d sector(s)",channel,address,sectors);
    if (buffer->mode!=CM_FILE || !buffer->file) {
        fprintf(lf," not writeable.\n");
        crc_clear();
        if (io_ops->sendb(0x80+25)) return 1;
        if (arguments.mode == M_ETHERNET && io_ops->sendb(0x80+25)) return 1;
        if (send_trailer(usecrc)) return 1;
        return 0;
    }
    if (!buffer->data) {
        buffer->data=malloc(65536);
        if (!buffer->data) exit(1);//out of memory?!
    }
    fseek(buffer->file, address << 8, SEEK_SET);
    if (arguments.mode == M_ETHERNET) {
        if (sectors>2) sectors=2;
        if (io_ops->getbytes(512*sectors, buffer->data)) return 1;
        crc_clear();
        if (fwrite(buffer->data,1,512*sectors,buffer->file)==512*sectors) {
            if (io_ops->sendb(0x80)) return 1;
            if (io_ops->sendb(0x80)) return 1;
        } else {
            if (io_ops->sendb(0x80+25)) return 1;
            if (io_ops->sendb(0x80+25)) return 1;
        }
        if (send_trailer(usecrc)) return 1;
    } else {
        crc_clear();
        if (io_ops->sendb(0x80)) return 1;
        if (send_trailer(usecrc)) return 1;

        gettimeofday(&start,NULL);
        while (sectors) {
    	    crc_clear();
  	    if (io_ops->getbytes(512, buffer->data)) return 1;
            if (usecrc) {
    	        if (io_ops->getb() < 0) return 1;
    	        if (crc_get()) {
    	    	    seterror(23,1);
    	    	    fprintf(lf," CRC error while write\n");
    	    	    io_ops->getb();
    	    	    return 1;
    	    	}
	    }
            fr = io_ops->getb();if (fr < 0) return 1;
            if (fr != 0x5a) {
                fprintf(lf, " frame error %02x\n",fr);
                return 1;
            }
    	    crc_clear();
    	    if (fwrite(buffer->data,1,512,buffer->file)==512) {
                if (sectors==1 || arguments.mode == M_RS232 || arguments.mode == M_RS232S) {
		    if (io_ops->sendb(0x80)) return 1;
            	    if (send_trailer(usecrc)) return 1;
            	}
    	    } else {
	        if (io_ops->sendb(0x80+25)) return 1;
    	        if (send_trailer(usecrc)) return 1;
    	        break;
    	    }
            sectors--;l+=512;
        }
	gettimeofday(&end,NULL);
        l2=((end.tv_sec-start.tv_sec)*1000000+end.tv_usec-start.tv_usec);
        if (l2) {
	    l=l*1000000/l2;
    	    fprintf(lf," %lld bytes/seconds\n",l);
        }
    }
    return 0;
}

#ifndef WIN32
static void testarg(int argc,char *argv[]) {
    int c;

    for (;;) {
        static struct option long_options[]= {
            {"user", required_argument, 0, 'u'},
            {"group", required_argument, 0, 'g'},
            {"root", required_argument, 0, 'r'},
            {"log", required_argument, 0, 'l'},
            {"nice", required_argument, 0, 'n'},
            {"allprg", no_argument, 0, 'P'},
            {"help", no_argument, 0, '?'},
            {"usage", no_argument, 0, 1},
            {"version", no_argument, 0, 'V'},
            {"background", no_argument, 0, 'b'},
            {"hog", no_argument, 0, 'h'},
            {"debug", no_argument, 0, 'D'},
            {"device", required_argument, 0, 'd'},
            {"mode", required_argument, 0, 'm'},
            {"lptport", required_argument, 0, 'p'},
            {"ipaddress", required_argument, 0, 'i'},
            {"network", required_argument, 0, 'N'},
            {0,0,0,0}
        };
        int option_index=0;

        c=getopt_long(argc, argv,
                      "m:u:g:r:l:n:P?VbhDd:p:i:N:"
                      , long_options, &option_index);
        if (c==-1) {
            if (optind==argc)
                break;
            c='?';
        }

        switch (c) {
        case '?':
            printf(
                   "Usage: ideservd [OPTION...]\n"
                   "IDEDOS 0.91 PCLink fileserver\n"
                   "\n"
                   "  -b, --background\t     Fork into background\n"
                   "  -d, --device=DEVICE\t     Device (/dev/parport0 or /dev/ttyS0)\n"
                   "  -g, --group=GROUP\t     Group under we run (nogroup)\n"
                   "  -h, --hog\t\t     CPU hog (off)\n"
                   "  -i, --ipaddress=IP\t     IP address of C64 on network\n"
                   "  -l, --log=FILE\t     Logfile (stdout)\n"
                   "  -m, --mode=MODE\t     Mode (x1541, xe1541, xm1541, xa1541,\n"
                   "\t\t\t     pc64, pc64s, rs232, rs232s usb, fifo, eth)\n"
                   "  -n, --nice=ADJUST\t     Adjust nice level (-19)\n"
                   "  -N, --network=NUM\t     Network number (0)\n"
                   "  -p, --lptport=IOPORT\t     Printer port address (0x378)\n"
                   "  -P, --allprg\t\t     Almost everything to PRG\n"
                   "  -r, --root=DIRECTORY\t     Root directory (.)\n"
                   "  -u, --user=USER\t     User under we run (nobody)\n"
                   "  -?, --help\t\t     Give this help list\n"
                   "      --usage\t\t     Give a short usage message\n"
                   "  -V, --version\t\t     Print program version\n"
                   "\n"
                   "Mandatory or optional arguments to long options are also mandatory or optional\n"
                   "for any corresponding short options.\n"
                   "\n"
                   "Report bugs to <soci@c64.rulez.org>.\n"
                   );
                   exit(0);
        case 1:
            printf(
                   "Usage: ideservd [-bhP?V] [-d DEVICE] [-g GROUP] [-i IP] [-l FILE] [-m MODE]\n"
                   "            [-n ADJUST] [-N NUM] [-p IOPORT] [-r DIRECTORY] [-u USER]\n"
                   "            [--background] [--device DEVICE] [--group=GROUP] [--hog]\n"
                   "            [--ipaddress IP] [--log=FILE] [--mode MODE] [--nice=ADJUST]\n"
                   "            [--network NUM] [--lptport=IOPORT] [--allprg] [--root=DIRECTORY]\n"
                   "            [--user=USER] [--help] [--usage] [--version]\n"
                  );
            exit(0);
        case 'V':
            printf("%s\n",argp_program_version);
            exit(0);
        case 'b':arguments.background=1;break;
        case 'P':arguments.relaxfiletypes=1;break;
        case 'D':arguments.debug=1;break;
        case 'u':arguments.user=optarg;break;
        case 'g':arguments.group=optarg;break;
        case 'l':arguments.log=optarg;break;
        case 'r':arguments.root=optarg;break;
        case 'h':arguments.hog=1;break;
        case 'd':arguments.device=optarg;break;
        case 'm':arguments.mode_name=optarg;break;
        case 'p':arguments.lptport=strtol(optarg,NULL,0) & 0xffff;break;
        case 'i':arguments.sin_addr=optarg;break;
        case 'N':arguments.network=strtol(optarg,NULL,0) & 0xff;break;
        case 'n':arguments.priority=strtol(optarg,NULL,0);break;
        default:
            exit(1);
        }
    }
}
#else // WIN32
void testarg(int argc,char *argv[]) {
    int j;
    for (j=1;j<argc;j++)
    {
	if (!strcmp(argv[j],"-?") || !strcmp(argv[j],"--help")) {
                MessageBox (NULL,
		"Usage: ideservd [OPTION...]\n"
		"IDEDOS 0.91 PCLink fileserver\n"
                "\n"
                "  -d, --device\t     Serial port device (COM1)\n"
                "  -h, --hog\t     CPU hog (off)\n"
                "  -i, --ipaddress\t     IP address of C64 on network\n"
                "  -l, --log\t\t     Logfile (ideservd.log)\n"
                "  -m, --mode\t     Mode (x1541, xe1541, xm1541, xa1541,\n"
                "\t\t     pc64, pc64s, rs232, rs232s usb, eth)\n"
                "  -N, --network\t     Network number (0)\n"
                "  -p, --lptport\t     Printer port address (0x378)\n"
                "  -P, --allprg\t     Almost everything to PRG\n"
                "  -r, --root=DIR\t     Root directory (.)\n"
		"  -?, --help\t     Give this help list\n"
		"      --usage\t     Give a short usage message\n"
		"  -V, --version\t     Print program version\n"
		"\n"
		"Mandatory or optional arguments to long options are also mandatory or optional\n"
		"for any corresponding short options.\n"
		"\n"
		"Report bugs to <soci@c64.rulez.org>.\n"
		,"Ideservd usage",MB_OK);
	    exit(1);
	}
        if (!strcmp(argv[j],"--usage")) {
            MessageBox(NULL,"Usage: ideservd [-hP?V] [-d DEVICE] [-i IP] [-l FILE] [-N NUM] [-p IOPORT] [-m MODE]\n"
                       "\t    [--device DEVICE] [--hog] [--ipaddress IP] [--log FILE] [--mode MODE] [--network NUM]\n"
                       "\t    [--lptport IOPORT] [--allprg] [--help] [--usage] [--version]\n" ,"Ideservd usage",MB_OK);
	    exit(1);
	}
	if (!strcmp(argv[j],"-V") || !strcmp(argv[j],"--version")) {
	    MessageBox(NULL,argp_program_version,"Ideservd version",MB_OK);
	    exit(1);
	}
	if (!strcmp(argv[j],"-P") || !strcmp(argv[j],"--allprg")) {arguments.relaxfiletypes=1;continue;}
	if (!strcmp(argv[j],"-l") || !strcmp(argv[j],"--log")) {
	    j++;if (j>=argc) goto ide2;
	    arguments.log=realloc(arguments.log, strlen(argv[j])+1);
	    strcpy(arguments.log,argv[j]);
	    continue;
        }
	if (!strcmp(argv[j],"-h") || !strcmp(argv[j],"--hog")) {arguments.hog=1;continue;}
	if (!strcmp(argv[j],"-D") || !strcmp(argv[j],"--debug")) {arguments.debug=1;continue;}
	if (!strcmp(argv[j],"-p") || !strcmp(argv[j],"--lptport")) {
	    j++;if (j>=argc) goto ide2;
	    arguments.lptport=strtol(argv[j],NULL,0) & 0xffff;
	    continue;
	}
	if (!strcmp(argv[j],"-d") || !strcmp(argv[j],"--device")) {
	    j++;if (j>=argc) goto ide2;
	    arguments.device=realloc(arguments.device, strlen(argv[j])+1);
	    strcpy(arguments.device,argv[j]);
	    continue;
        }
	if (!strcmp(argv[j],"-i") || !strcmp(argv[j],"--ipaddress")) {
	    j++;if (j>=argc) goto ide2;
	    arguments.sin_addr=realloc(arguments.sin_addr, strlen(argv[j])+1);
	    strcpy(arguments.sin_addr,argv[j]);
	    continue;
	}
	if (!strcmp(argv[j],"-m") || !strcmp(argv[j],"--mode")) {
	    j++;if (j>=argc) goto ide2;

	    arguments.mode_name=realloc(arguments.mode_name, strlen(argv[j])+1);
	    strcpy(arguments.mode_name,argv[j]);
	    continue;
        }
	if (!strcmp(argv[j],"-N") || !strcmp(argv[j],"--network")) {
	    j++;if (j>=argc) goto ide2;
	    arguments.network=strtol(argv[j],NULL,0) & 0xff;
	    continue;
	}
	if (!strcmp(argv[j],"-r") || !strcmp(argv[j],"--root")) {
	    j++;if (j>=argc) goto ide2;
	    arguments.root=realloc(arguments.root, strlen(argv[j])+1);
	    strcpy(arguments.root,argv[j]);
	    continue;
        }

/*        if (!strcmp(argv[j],"-n")) {
	    int i=0;
	    j++;if (j>=argc)*/if (0) {
		ide2:
		MessageBox(NULL,"Ideservd: option requires an argument!",NULL,MB_ICONERROR|MB_OK);
		goto ide;
	    }
/*
	    while (argv[j][i] && argv[j][i]!='=') i++;
            if (argv[j][i]=='=') {
                argv[j][i]=0;
                arguments.priority=strtol(argv[j],NULL,0);
            }
            continue;
        }*/
	MessageBox(NULL,"Usage: ideservd [OPTION...]\n",NULL,MB_OK);
	ide:
        MessageBox(NULL,"Try `ideservd --help' or `ideservd --usage' for more information.\n",NULL,MB_OK);
        if (arguments.log) free(arguments.log);
        if (arguments.device) free(arguments.device);
        exit(1);
    }
}
#endif

static void init(int argc,char *argv[])
{
#ifndef WIN32
    struct passwd *user2;
    struct group *group2;
#endif
    char *init_dev = NULL, init_int = 0;
    arguments.mode=M_USB;
    arguments.lptport=DEFAULT_IOPORT;
    arguments.user = "nobody";
    arguments.group = "nogroup";
    arguments.priority = -19;
    testarg(argc, argv);
    partition_create(1, (c64c_t *)"PARTITION 1");
    partition_select(1);

#ifdef WIN32
    lf = (arguments.log && !strcmp(arguments.log, "-")) ? stdout : fopen(arguments.log ? arguments.log : "ideservd.log", "wt");
#else
    lf = arguments.log && strcmp(arguments.log, "-") ? fopen(arguments.log, "wt") : stdout;
#endif
    if (!lf) {
        lf = stdout;
        fprintf(lf, "Cannot open logfile '%s'\n", arguments.log);
    }
    fprintf(lf, "%s\n", argp_program_version);

    fprintf(lf, "Using driver ");

    arguments.mode = M_NONE;
    if (arguments.mode_name) {
        const struct {
            const char *name;
            enum e_modes mode;
        } *i, modes[] = {
#ifdef _X1541_H
            {"x1541", M_X1541}, {"xe1541", M_XE1541}, {"xm1541", M_XM1541}, {"xa1541", M_XA1541},
#endif
#ifdef _PC64_H
            {"pc64", M_PC64}, {"pc64s", M_PC64S},
#endif
#ifdef _RS232_H
            {"rs232", M_RS232}, {"rs232s", M_RS232S},
#endif
#ifdef _USB_H
            {"usb", M_USB},
#endif
#ifdef _FIFO_H
            {"fifo", M_FIFO},
#endif
#ifdef _ETH_H
            {"eth", M_ETHERNET},
#endif
            {NULL, M_NONE}
        };
        i = modes;
        while (i->name) {
            if (!strcasecmp(arguments.mode_name, i->name)) {
                arguments.mode = i->mode;
                break;
            }
            i++;
        }
    } else arguments.mode = M_USB;

    switch (arguments.mode) {
#ifdef _X1541_H
    case M_X1541:
    case M_XE1541:
    case M_XM1541:
    case M_XA1541:
        {
            int i;
            const char *cable = "X1541";
            io_ops=&x1541_io_ops;
            i=arguments.lptport & 0xffff;
            if (arguments.hog) i|=0x10000;
            if (arguments.device) i|=0x20000;
            if (arguments.mode==M_XE1541) {i |= 0x40000; cable = "XE1541";}
            if (arguments.mode==M_XM1541) {i |= 0x80000; cable = "XM1541";}
            if (arguments.mode==M_XA1541) {i |= 0xc0000; cable = "XA1541";}
            fprintf(lf, "%s using %s cable", io_ops->drivername(), cable);
            if (arguments.device) {
                fprintf(lf, " on device %s\n", arguments.device);
            } else { 
                fprintf(lf, " at port %x\n", arguments.lptport);
            }
            init_dev = arguments.device;
            init_int = i;
        }
        break;
#endif
#ifdef _PC64_H
    case M_PC64:
    case M_PC64S:
        {
            int i;
            io_ops=&pc64_io_ops;
            i = arguments.lptport & 0xffff;
            if (arguments.hog) i|=0x10000;
            if (arguments.device) i|=0x20000;
            if (arguments.mode==M_PC64S) i|=0x40000;
            if (arguments.device) {
                fprintf(lf, "%s on device %s\n", io_ops->drivername(), arguments.device);
            } else { 
                fprintf(lf, "%s at port %x\n", io_ops->drivername(), arguments.lptport);
            }
            init_dev = arguments.device;
            init_int = i;
        }
        break;
#endif
#ifdef _RS232_H
    case M_RS232:
    case M_RS232S:
        io_ops=&rs232_io_ops;
        if (!arguments.device) arguments.device = DEFAULT_COMPORT;
        fprintf(lf, "%s on device %s\n", io_ops->drivername(), arguments.device);
        init_dev = arguments.device;
        init_int = (arguments.mode == M_RS232S);
        break;
#endif
#ifdef _USB_H
    case M_USB:
        io_ops=&usb_io_ops;
        init_dev = arguments.device;
        fprintf(lf, "%s for device %s\n", io_ops->drivername(), arguments.device ? arguments.device : "*");
        break;
#endif
#ifdef _FIFO_H
    case M_FIFO:
        io_ops=&fifo_io_ops;
        fprintf(lf, "%s\n", io_ops->drivername());
        break;
#endif
#ifdef _ETH_H
    case M_ETHERNET:
        io_ops = &eth_io_ops;
        init_dev = arguments.sin_addr;
        init_int = arguments.network;
        fprintf(lf, "%s to %s on network %d\n", io_ops->drivername(), arguments.sin_addr ? arguments.sin_addr : "0.0.0.0", arguments.network);
        break;
#endif
    default: 
        fprintf(lf, "NONE for unknown mode '%s'\n", arguments.mode_name);
        exit(1);
    }

#ifndef WIN32
    {
        pid_t pid;

        if (arguments.background) if (fork()) {fprintf(lf, "Forking into background.\n");exit(0);}
        for (;;) {
            pid=fork();
            if (pid) {
                usleep(1000000);
                waitpid(pid, NULL, 0);
            } else break;
        }
    }
#endif

    if (io_ops->initialize(init_dev, init_int, lf)) exit(1);

#ifdef WIN32
    if (arguments.root) chdir(arguments.root); else chdir("/cygdrive");
#else
    nice(arguments.priority);
    if (!(user2=getpwnam(arguments.user))) {fprintf(lf, "User %s does not exists.\n",arguments.user);exit(1);}
    if (!(group2=getgrnam(arguments.group))) {fprintf(lf, "Group %s does not exists.\n",arguments.group);exit(1);}
// chroot crashes on WIN32 with current cygwin1.dll without registry entry:
// [HKEY_LOCAL_MACHINE\SOFTWARE\Cygnus Solutions\Cygwin\mounts v2\/]
// "native"="C:\\"
    if (arguments.root) chdir(arguments.root);
    if (arguments.mode != M_FIFO) {
        if (arguments.root) if (chroot(arguments.root) || chdir("/")) {fprintf(lf, "Changing root dir to '%s'\n",arguments.root);exit(1);}
        if (setgid(group2->gr_gid) || setuid(user2->pw_uid)) {fprintf(lf, "Could not drop privileges.\n");exit(1);}
    }
    fclose(stdin);
    fclose(stderr);
    if (lf != stdout) fclose(stdout);
#endif
}

#ifdef WIN32
INT_PTR CALLBACK MyDlgProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
/*  int i;
  HWND hwnd_select;

  const char *pclinkmodes[] = {
            "x1541",
            "xe1541",
            "xm1541",
            "xa1541",
            "pc64",
            "pc64s",
            "rs232",
            "rs232s",
            "usb",
            "eth" };*/

  switch(message) 
  {
//   case WM_INITDIALOG:
//     hwnd_select = GetDlgItem(hwnd, IDC_SELECT);
//     for(i=0; i<11; i++) SendMessage(hwnd_select, CB_ADDSTRING, 0, (LPCTSTR)pclinkmodes[i]);
//     SendMessage(hwnd_select, CB_SETCURSEL, 8, 0); // from 0
//     return TRUE;
//   break;

   case WM_COMMAND:
	if (LOWORD(wParam) == IDC_EXIT || LOWORD(wParam) == IDCANCEL)
	{
		PostQuitMessage(0);
		EndDialog(hwnd,TRUE);
		return TRUE;
        }

	if (LOWORD(wParam) == IDC_BUTTON)
	{
	  ShowWindow(hwnd, SW_HIDE);
	}
   break;

//   case WM_SYSCOMMAND:
//	if (LOWORD(wParam) == SC_MINIMIZE)
//	{
//	   ShowWindow(hwnd, SW_HIDE);
//	}
//   break;

   case WM_USER:
	if (lParam == WM_RBUTTONDOWN || lParam == WM_LBUTTONDOWN)
        {
	   ShowWindow(hwnd, SW_SHOW);
   	}
   break;
  }
return FALSE;
}
#endif

int main(int argc,char *argv[])
// int WINAPI WinMain (HINSTANCE hInstance,
//                    HINSTANCE hPrevInstance,
//                    LPSTR szCmdLine,
//                    int iCmdShow)
{
    int b;
    init(argc,argv);
    signal(SIGALRM,timeout);
    signal(SIGTERM,terminate);
    signal(SIGINT,terminate);
    for (b=0;b<16;b++) {buff[b].mode=CM_CLOSED;buff[b].data=NULL;buff[b].file=NULL;}

#ifdef WIN32
    HWND hwnd;
    MSG msg;

    hwnd=CreateDialog(NULL, MAKEINTRESOURCE(IDD_MAIN), NULL, (DLGPROC)MyDlgProc);
    ShowWindow(hwnd, SW_SHOW);
 
// TRAY ICON

    memset(&g_notifyIconData, 0, sizeof(NOTIFYICONDATA));
    g_notifyIconData.cbSize = sizeof(NOTIFYICONDATA);
    g_notifyIconData.hWnd = hwnd;
    g_notifyIconData.uID = 0;
    g_notifyIconData.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
    g_notifyIconData.uCallbackMessage = WM_USER; 
    g_notifyIconData.hIcon = (HICON)LoadImage( NULL, TEXT("pclink.ico"), IMAGE_ICON, 0, 0, LR_LOADFROMFILE);
    strcpy(g_notifyIconData.szTip, "IDE64 PCLink");
    Shell_NotifyIcon(NIM_ADD, &g_notifyIconData);
    
    HANDLE hThread;
    DWORD iID;
    hThread = CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)PCLinkLoop,NULL,0,&iID);
    
    while(GetMessage (&msg,NULL,0,0)!=0) {
//       TranslateMessage(&msg);  // Only if there is text input
       DispatchMessage(&msg);
    }
    TerminateThread(hThread,0);
    Shell_NotifyIcon(NIM_DELETE, &g_notifyIconData);
    terminate();
    return 0;
}

long WINAPI PCLinkLoop(long lParam)
{
    int b;
#endif //WIN32

    for (;;) {
        fprintf(lf,"ready.\n");
        fflush(lf);
    start2:
        crc_clear();
        b=io_ops->wait();
        if (b<0) {
#ifdef WIN32
            usleep(1000000);
            io_ops->reinitialize(lf);
            b=0;
#else
            terminate();	//@@@@@
#endif
        }

        switch (b) {
        case 0x00:goto start2;
        case 'N':b=openfile(0);break;
        case 0xCE:b=openfile(1);break;
        case 'G':b=readfile(0);break;
        case 0xC7:b=readfile(1);break;
        case 'S':b=writefile(0);break;
        case 0xD3:b=writefile(1);break;
        case 'D':b=closefile(0);break;
        case 0xC4:b=closefile(1);break;
        case 'I':b=statuserror(0);break;
        case 0xC9:b=statuserror(1);break;
        case 'O':b=openfile_compat();break;
        case 'R':b=readfile_compat();break;
        case 'C':b=closefile_compat();break;
        default:fprintf(lf,"Unknown command %02X\n",b);
        }
        if (b) io_ops->clean();
    }
    return 0;
}
